/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EXTENDJOBMAPPER_CPP
#define EXTENDJOBMAPPER_CPP

#include "ExtendJobMapper.h"
#include <typeinfo>
#include <cstring>
#include <omp.h>
#include "ScoredSeqSubseq.h"
#include "ScoredSeqMonoScore.h"
using namespace::std;


ExtendJobMapper::~ExtendJobMapper(){}



ExtendJobMapperWithLinks::ExtendJobMapperWithLinks(){}

ExtendJobMapperWithLinks::ExtendJobMapperWithLinks(set<ScoredSeqWithCarriers*>* contigsWithCarriers, float fractMapId) :
  _fractMapId(fractMapId){
  _edgeSize = 0;
  for (set<ScoredSeqWithCarriers*>::iterator it = contigsWithCarriers->begin(); it != contigsWithCarriers->end(); ++it){
    if ( (*it)->size() > _edgeSize ){ _edgeSize = (*it)->size(); }
  }
  constructorHelper(contigsWithCarriers);  
}
ExtendJobMapperWithLinks::ExtendJobMapperWithLinks(set<ScoredSeqWithCarriers*>* contigsWithCarriers, float fractMapId, long edgeSize) :
  _edgeSize(edgeSize),
  _fractMapId(fractMapId){
  constructorHelper(contigsWithCarriers);  
}
void ExtendJobMapperWithLinks::constructorHelper(set<ScoredSeqWithCarriers*>* contigsWithCarriers){

  _shortArray = new ScoredSeq*[ contigsWithCarriers->size() * 2 + 1 ];
  _flipArray = new ScoredSeq*[ contigsWithCarriers->size() * 2 + 1 ];

  _plusIndex = ExtendCycle::plusIndex();
  _minusIndex = ExtendCycle::minusIndex();
  _frontIndex = ExtendCycle::frontIndex();
  _backIndex = ExtendCycle::backIndex();

  // use the contigsWithCarriers to construct the front and back BWT subseq collections
  // minContigSplitLen -> edgeSize

  // first the front edges
  set<ScoredSeq*> inputSet;
  _disposableIndex = 0;

  for (set<ScoredSeqWithCarriers*>::iterator it = contigsWithCarriers->begin(); it != contigsWithCarriers->end(); ++it){
    ScoredSeqWithCarriers* carrier = (*it);
    ScoredSeq* keySeqFront;
    ScoredSeq* keySeqBack;
    if ( _edgeSize >= 0 and carrier->size() >= _edgeSize ){
      keySeqFront = new ScoredSeqSubseq( carrier, 0, _edgeSize);
      keySeqBack = new ScoredSeqSubseq( carrier, carrier->size() - _edgeSize, _edgeSize);
    } else {
      keySeqFront = new ScoredSeqSubseq( carrier, 0, carrier->size() );
      keySeqBack = new ScoredSeqSubseq( carrier, 0, carrier->size());
    }
    ScoredSeq* inSeqFront = ScoredSeqFlip::getFlip(keySeqFront,'-');
    ScoredSeq* inSeqBack = ScoredSeqFlip::getFlip(keySeqBack,'+');
    inputSet.insert(inSeqFront);
    inputSet.insert(inSeqBack);

    _shortArray[_disposableIndex] = keySeqFront;
    _flipArray[_disposableIndex] = inSeqFront;
    ++_disposableIndex;
    _shortArray[_disposableIndex] = keySeqBack;
    _flipArray[_disposableIndex] = inSeqBack;
    ++_disposableIndex;

    keySeqFront->buffer();
    keySeqBack->buffer();
  }
  _mappingCollection = new ScoredSeqCollectionBwt(&inputSet, _fractMapId, false);
  _mappingCollection->disableEdgeScaling();

}



ExtendJobMapperWithLinks::~ExtendJobMapperWithLinks(){
  // get rid of the subseq keys, not the contig values
  for (long n = 0; n < _disposableIndex; ++n){
    delete _shortArray[n];
    delete _flipArray[n];
  }
  delete [] _shortArray;
  delete [] _flipArray;

  // and delete the collections themselves
  delete _mappingCollection;
}



long ExtendJobMapperWithLinks::mapQueries(ScoredSeqWithCarriers** queries, long numQueries){
  // save the ones with hits in this array; at the end, place them in the input array
  ScoredSeqWithCarriers** queriesWithHits = new ScoredSeqWithCarriers*[ numQueries + 1 ];
  long countWithHits = 0;

  // make a non-redundant sequence set
  map<string, vector<ScoredSeqWithCarriers*>*> seqToQuerySet;
  long numNr = 0;
  for (long n = 0; n < numQueries; ++n){
    char* querySeq = queries[n]->getSeq('+');
    map<string, vector<ScoredSeqWithCarriers*>*>::iterator foundSeqIt = seqToQuerySet.find( querySeq );
    if (foundSeqIt == seqToQuerySet.end()){
      vector<ScoredSeqWithCarriers*>* newSet = new vector<ScoredSeqWithCarriers*>;
      newSet->push_back(queries[n]);
      seqToQuerySet.insert( pair<string, vector<ScoredSeqWithCarriers*>*> (querySeq, newSet) );
    } else { foundSeqIt->second->push_back( queries[n] ); }
    queries[n]->deepUnbuffer();
    delete [] querySeq;
  }

  // now do the mappings for each query
  for (map<string, vector<ScoredSeqWithCarriers*>*>::iterator seqIt = seqToQuerySet.begin();
       seqIt != seqToQuerySet.end(); ++seqIt){
    // don't delete this, it will be deleted with 'query'
    ScoredSeq* query = new ScoredSeqMonoScore(seqIt->first, 1);

    // set up the collections
    vector<ScoredSeqWithCarriers*> matchingContigsPlus;
    vector<ScoredSeqWithCarriers*> matchingContigsMinus;
    findMatchesHelper(query, &matchingContigsPlus, &matchingContigsMinus);

    // if there are no matches at all, delete everyting
    if (matchingContigsPlus.size() + matchingContigsMinus.size() == 0){
      for (vector<ScoredSeqWithCarriers*>::iterator it = seqIt->second->begin(); it != seqIt->second->end(); ++it){
	(*it)->deepUnbuffer();
	(*it)->deepDelete();
      }
    } else {
      // schedule the SSN versions of the reads for ultimate destruction
      for (vector<ScoredSeqWithCarriers*>::iterator it = seqIt->second->begin(); it != seqIt->second->end(); ++it){
	queriesWithHits[ countWithHits ] = (*it);
	countWithHits++;
	(*it)->deepUnbuffer();
      }

      // sort out the matches with the appropriate binning
      if (matchingContigsPlus.size() > 0){
	for (vector<ScoredSeqWithCarriers*>::iterator it = seqIt->second->begin(); it != seqIt->second->end(); ++it){
	  addToContigMatchesHelper( *it, _plusIndex, _backIndex, &matchingContigsPlus);
	}
      }
      if (matchingContigsMinus.size() > 0){
	for (vector<ScoredSeqWithCarriers*>::iterator it = seqIt->second->begin(); it != seqIt->second->end(); ++it){
	  addToContigMatchesHelper( *it, _minusIndex, _frontIndex, &matchingContigsMinus);
	}
      }
    }
    delete query;
    delete seqIt->second;
  }

  for (long n = 0; n < countWithHits; ++n){ queries[n] = queriesWithHits[n]; }
  delete [] queriesWithHits;
  return countWithHits;
}


// MODIFIES: matchingContigs AND alreadyQueriedToHits
void ExtendJobMapperWithLinks::findMatchesHelper(ScoredSeq* currentQuery,
						 vector<ScoredSeqWithCarriers*>* matchingContigsPlus,
						 vector<ScoredSeqWithCarriers*>* matchingContigsMinus){
  vector<Alignment*> matchAlignments;
  _mappingCollection->getBestMatches( &matchAlignments, currentQuery, '+', currentQuery->size(), ScoredSeqCollectionBwt::softMinOvl);

  for (vector<Alignment*>::iterator alIt = matchAlignments.begin(); alIt != matchAlignments.end(); ++alIt){
    // find the full contig
    ScoredSeqFlip* flipContig = dynamic_cast<ScoredSeqFlip*>( (*alIt)->seqB() );
    ScoredSeqSubseq* partialContig = dynamic_cast<ScoredSeqSubseq*>( flipContig->getNested() );
    ScoredSeqWithCarriers* fullContig = dynamic_cast<ScoredSeqWithCarriers*>( partialContig->getNested() );
    switch ( flipContig->getSense() ){
    case '+': matchingContigsPlus->push_back( fullContig ); break;
    case '-': matchingContigsMinus->push_back( fullContig ); break;
    default: throw AssemblyException::LogicError("EJMwLinks::findMatchesHelper got a bad sense char");
    }
    delete (*alIt);
  }
}


void ExtendJobMapperWithLinks::addToContigMatchesHelper(ScoredSeqWithCarriers* seqToAdd, int side, int frontOrBack, vector<ScoredSeqWithCarriers*>* matchingContigs){
  #pragma omp critical (EJMlink)
  {
    for (vector<ScoredSeqWithCarriers*>::iterator mcIt = matchingContigs->begin(); mcIt != matchingContigs->end(); ++mcIt){
      (*mcIt)->addCarriedSeq( seqToAdd, frontOrBack );
      seqToAdd->addCarriedSeq( (*mcIt), side );
    }
  }
}







ExtendJobMapperNoLinks::ExtendJobMapperNoLinks(){}

ExtendJobMapperNoLinks::ExtendJobMapperNoLinks(set<ScoredSeqWithCarriers*>* contigsWithCarriers, float fractMapId) :
  _fractMapId(fractMapId){
  _edgeSize = 0;
  for (set<ScoredSeqWithCarriers*>::iterator it = contigsWithCarriers->begin(); it != contigsWithCarriers->end(); ++it){
    if ( (*it)->size() > _edgeSize ){ _edgeSize = (*it)->size(); }
  }
  constructorHelper(contigsWithCarriers);  
}
ExtendJobMapperNoLinks::ExtendJobMapperNoLinks(set<ScoredSeqWithCarriers*>* contigsWithCarriers, float fractMapId, long edgeSize) :
  _edgeSize(edgeSize),
  _fractMapId(fractMapId){
  constructorHelper(contigsWithCarriers);  
}
void ExtendJobMapperNoLinks::constructorHelper(set<ScoredSeqWithCarriers*>* contigsWithCarriers){

  // use the contigsWithCarriers to construct the front and back BWT subseq collection (one collection)
  set<ScoredSeq*> shortSet;

  // first the front edges
  for (set<ScoredSeqWithCarriers*>::iterator it = contigsWithCarriers->begin(); it != contigsWithCarriers->end(); ++it){
    ScoredSeqWithCarriers* carrier = (*it);
    ScoredSeq* keySeq;
    if ( _edgeSize >= 0 and carrier->size() >= _edgeSize ){
      keySeq = new ScoredSeqSubseq( carrier, 0, _edgeSize);
    } else {
      keySeq = new ScoredSeqSubseq( carrier, 0, carrier->size() );
    }
    shortSet.insert(keySeq);
    keySeq->buffer();
  }

  // now the back edges
  set<ScoredSeq*> shortBackSet;
  for (set<ScoredSeqWithCarriers*>::iterator it = contigsWithCarriers->begin(); it != contigsWithCarriers->end(); ++it){
    ScoredSeqWithCarriers* carrier = (*it);
    ScoredSeq* keySeq;
    if ( _edgeSize >= 0 and carrier->size() >= _edgeSize ){
      keySeq = new ScoredSeqSubseq( carrier, carrier->size() - _edgeSize, _edgeSize);
    } else {
      keySeq = new ScoredSeqSubseq( carrier, 0, carrier->size());
    }
    shortSet.insert(keySeq);
    keySeq->buffer();
  }

  _mappingCollection = new ScoredSeqCollectionBwt(&shortSet, _fractMapId, false);
  _mappingCollection->disableEdgeScaling();

  // MORE HERE???
}



ExtendJobMapperNoLinks::~ExtendJobMapperNoLinks(){
  // get rid of the subseq keys, not the contig values
  set<ScoredSeq*> outerToDelete;
  _mappingCollection->getSeqs( &outerToDelete );
  for (set<ScoredSeq*>::iterator it = outerToDelete.begin(); it != outerToDelete.end(); ++it){ delete (*it); }
  delete _mappingCollection;
}


ExtendJobMapper::QueryCarrier::QueryCarrier(char* query, long size) : _query(query), _size(size) {}
ExtendJobMapper::QueryCarrier::~QueryCarrier(){}


long ExtendJobMapperNoLinks::mapQueries(ScoredSeqWithCarriers** queries, long numQueries){
  // save the ones with hits in this array; at the end, place them in the input array
  ScoredSeqWithCarriers** queriesWithHits = new ScoredSeqWithCarriers*[ numQueries + 1 ];
  long countWithHits = 0;

  // make a non-redundant sequence set
  map<string, QueryCarrier*> seqToQuerySet;
  long numNr = 0;
  for (long n = 0; n < numQueries; ++n){
    char* querySeq = queries[n]->getSeq('+');
    map<string, QueryCarrier*>::iterator foundSeqIt = seqToQuerySet.find( querySeq );
    if (foundSeqIt == seqToQuerySet.end()){
      QueryCarrier* newCarrier = new QueryCarrier(querySeq, queries[n]->size());
      newCarrier->_seqs.push_back(queries[n]);
      seqToQuerySet.insert( pair<string, QueryCarrier*> (querySeq, newCarrier) );
    } else {
      foundSeqIt->second->_seqs.push_back( queries[n] );
      delete [] querySeq;
    }
    queries[n]->deepUnbuffer();
  }

  // set up the arrays
  long numMono = seqToQuerySet.size();
  vector<ScoredSeqWithCarriers*>** likeSequences = new vector<ScoredSeqWithCarriers*>*[ numMono+1 ];
  ScoredSeq** monoQueries = new ScoredSeq*[ numMono+1 ];
  long* queryLengths = new long[ numMono+1 ];
  QueryCarrier** carrierArray = new QueryCarrier*[ numMono+1 ];
  // now use this as an index iterator - it will return to the full value
  numMono = 0;
  for (map<string, QueryCarrier*>::iterator seqIt = seqToQuerySet.begin();
       seqIt != seqToQuerySet.end(); ++seqIt){
    likeSequences[numMono] = &(seqIt->second->_seqs);
    monoQueries[numMono] = new ScoredSeqMonoScore(seqIt->second->_query, 1, seqIt->second->_size, true);
    queryLengths[numMono] = seqIt->second->_size;
    carrierArray[numMono] = seqIt->second;
    ++numMono;
  }

  bool* qHasMatch = _mappingCollection->hasMatch(numMono, monoQueries, queryLengths,
						 ScoredSeqCollectionBwt::notThreaded,
						 ScoredSeqCollectionBwt::softMinOvl );

  for (long mqN = 0; mqN < numMono; ++mqN){
    if ( qHasMatch[mqN] ){
      for (vector<ScoredSeqWithCarriers*>::iterator it = likeSequences[mqN]->begin(); it != likeSequences[mqN]->end(); ++it){
	queriesWithHits[ countWithHits ] = (*it);
	countWithHits++;
	(*it)->deepUnbuffer();
      }
    } else {
      for (vector<ScoredSeqWithCarriers*>::iterator it = likeSequences[mqN]->begin(); it != likeSequences[mqN]->end(); ++it){
	(*it)->deepUnbuffer();
	(*it)->deepDelete();
      }
    }
    delete monoQueries[mqN];
    //delete likeSequences[mqN];
    delete carrierArray[mqN];
  }

  delete [] monoQueries;
  delete [] likeSequences;
  delete [] carrierArray;
  delete [] queryLengths;
  delete [] qHasMatch;

  for (long n = 0; n < countWithHits; ++n){ queries[n] = queriesWithHits[n]; }
  delete [] queriesWithHits;
  return countWithHits;
}







#endif
